;; Generated package description from qrencode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "qrencode" "1.4beta1.0.20260725.0" "QRCode encoder" '((emacs "25.1")) :commit "d96e458305c07ce845f3ebff573416972d1f8834" :authors '(("Rüdiger Sonderfeld" . "ruediger@c-plusplus.net")) :maintainer '("Rüdiger Sonderfeld" . "ruediger@c-plusplus.net") :keywords '("qrcode" "comm") :url "https://github.com/ruediger/qrencode-el")
